/* $Id$ */
/*
 * simplified drawing board for X window
 */
#include <X11/Xlib.h>
#include <X11/Xatom.h>      /* XA_STRING */
#include <X11/Xproto.h>     /* CARD8, CARD32 */
#include <X11/Xutil.h>      /* XTextProperty */
#include <X11/cursorfont.h> /* XC_clock */
#include <pwd.h>            /* getpwnam(), getpwuid() */
#include <unistd.h>         /* getuid() */
#include "common.h"
#include "dboard.h"

#define PROGNAME "dummy"
#define TITLE_DEFAULT "Title sample"

static unsigned long dBoard_get_background(DBoard *board);
static void dBoard_get_window_pos(DBoard *board, XWindowAttributes *xwa);
static void dBoard_set_window_pos(DBoard *board, XWindowAttributes *xwa);

static void dBoard_init_cursor(DBoard *board);
static void dBoard_free_cursor(DBoard *board);

static void dBoard_init_color(DBoard *board);
static int dBoard_pending(DBoard *board);
static int dBoard_nextEvent(DBoard *board, XEvent *event);
static XrmDatabase dBoard_get_xrmdb(Display *d, Window window_id);
static int dBoard_preparse_options(char **name, char **display, int *argc, char **argv);
static void dBoard_parse_options(DBoard *board, int *argc, char **argv);
static int dBoard_resource(DBoard *board, char *progname, char *name, char *class, XrmValue *value);

static  XrmOptionDescRec options[] = {
  {"-iconic",      ".iconic",              XrmoptionNoArg,         "on"},
  {"-name",        ".name",                XrmoptionSepArg,        NULL},
  {"-title",       ".title",               XrmoptionSepArg,        NULL},
  {"-display",     ".display",             XrmoptionSepArg,        NULL},
  {"-geometry",    ".geometry",            XrmoptionSepArg,        NULL},
  {"-fg",          "*foreground",          XrmoptionSepArg,        NULL},
  {"-bg",          "*background",          XrmoptionSepArg,        NULL},
  {"-foreground",  "*foreground",          XrmoptionSepArg,        NULL},
  {"-background",  "*background",          XrmoptionSepArg,        NULL},
  {"-borderwidth", ".borderWidth",         XrmoptionSepArg,        NULL},
  {"-xrm",         NULL,                   XrmoptionResArg,        NULL},
};

/*
 * Open DBoard.
 */
DBoard *dBoard_open(int *argc, char **argv)
{
  DBoard *board;
  Display *d;
  Window w;
  XrmDatabase res;
  char *name, *display;
  int s; /* screen number */
  XrmValue value;
  int xoff, yoff;
  unsigned int width, height, border;
  unsigned int width_max, height_max;
  XSizeHints hints;

  /* initialize X resources */
  XrmInitialize();
  name = display = NULL;
  dBoard_preparse_options(&name, &display, argc, argv);
  d = XOpenDisplay(display);
  if (d == NULL) {
    return NULL;
  }
  res = dBoard_get_xrmdb(d, 0);
  board = XMALLOC(DBoard, 1);
  board->display = d;
  board->resource = res;
  dBoard_parse_options(board, argc, argv);

  /* setup parameters */
  border = 2;
  if (dBoard_resource(board, PROGNAME, ".borderWidth", NULL, &value)) {
    sscanf((char *)value.addr, "%u", &border);
  }
  s = XDefaultScreen(d);
  width_max = XDisplayWidth(d, s);
  height_max = XDisplayHeight(d, s);
  width = 100;
  height = 200;
  xoff = 30;
  yoff = 40;
  hints.flags = 0;
  hints.win_gravity = NorthWestGravity;
  if (dBoard_resource(board, PROGNAME, ".geometry", NULL, &value)) {
    int rc;
    rc = XParseGeometry((char *)value.addr, &xoff, &yoff, &width, &height);
    DEBUG_PRINT("before:xoff=%d, yoff=%d, width=%d, height=%d, border=%d\n", xoff, yoff, width, height, border);
    if ((rc & XValue) || (rc & YValue)) {
      hints.flags = USPosition;
    }
    if ((rc & XValue) && (rc & XNegative)) {
      hints.win_gravity = NorthEastGravity;
      xoff = width_max-(width+2*border)-abs(xoff);
    }
    if ((rc & YValue) && (rc & YNegative)) {
      hints.win_gravity = (hints.win_gravity == NorthWestGravity) ?
        SouthWestGravity : SouthEastGravity;
      yoff = height_max-(height+2*border)-abs(yoff);
    }
    DEBUG_PRINT("after:xoff=%d, yoff=%d, width=%d, height=%d, border=%d\n", xoff, yoff, width, height, border);
  }
  unsigned long background = dBoard_get_background(board);
  w = XCreateSimpleWindow(d, RootWindow(d, s), xoff, yoff, width, height,
                          border, WhitePixel(d, s), background);
  board->window = w;

  char *title = TITLE_DEFAULT;
  if (dBoard_resource(board, PROGNAME, ".title", NULL, &value)) {
    title = (char *)value.addr;
  }

  XTextProperty w_name;

  w_name.value = (unsigned char *)title;
  w_name.encoding = XA_STRING;
  w_name.format = 8;
  w_name.nitems = strlen(title);
  hints.x = xoff;
  hints.y = yoff;
  hints.width = width;
  hints.height = height;
  hints.max_width = width_max*2;
  hints.max_height = height_max*2;
  hints.flags |= USSize | PMaxSize | PWinGravity;
  XSetWMProperties(d, w, &w_name, NULL, argv, *argc, &hints, NULL, NULL);
#if 0
  XmbSetWMProperties(d, w, title, NULL, argv, *argc, &hints, NULL, NULL);
  Xutf8SetWMProperties(d, w, title, NULL, argv, *argc, &hints, NULL, NULL);
#endif /* 0 */
  GC gc = XCreateGC(d, w, 0, 0); /* create graphics contexts */
  board->gc = gc;
  board->pixmap = None;
  dBoard_init_color(board);
  dBoard_init_cursor(board);
  return board;
}

/*
 * window information.
 */
void dBoard_get_attributes(DBoard *board, int *depth, int *width, int *height, unsigned long *background)
{
  Display *display;
  int screen;

  display = board->display;
  screen = XDefaultScreen(display);
  *depth = XDefaultDepth(display, screen);
  *width = XDisplayWidth(display, screen);
  *height = XDisplayHeight(display, screen);
  *background = dBoard_get_background(board);
}

static unsigned long dBoard_get_background(DBoard *board)
{
  unsigned long background;
  XrmValue value;
  Display *d;
  int s;

  d = board->display;
  s = XDefaultScreen(d);
  background = BlackPixel(d, s);
  if (dBoard_resource(board, PROGNAME, "*background", NULL, &value)) {
    XColor screen_def, exact_def;
    fprintf(stderr, "background = %s\n", (char *)value.addr);
    if (XAllocNamedColor(d, XDefaultColormap(d, s), (char *)value.addr, &screen_def, &exact_def)) {
      background = screen_def.pixel;
    }
  }
  return background;
}

/*
 * returns the x, y, w, h coords of window.
 * x, y are relative to root window the border is not included
 * (x, y map to top-left pixel in window)
 */
static void dBoard_get_window_pos(DBoard *board, XWindowAttributes *xwa)
{
  Display *d;
  Window w, child;
  int s;

  d = board->display;
  w = board->window;
  s = XDefaultScreen(d);
  /* get the window width and height */
  XGetWindowAttributes(d, w, xwa);
  /* get the window origin */
  XTranslateCoordinates(d, w, RootWindow(d, s), 0, 0, &xwa->x, &xwa->y, &child);
  DEBUG_PRINT("xwa:(%d, %d) %dx%d border=%d\n",
              xwa->x, xwa->y, xwa->width, xwa->height, xwa->border_width);
}

static void dBoard_set_window_pos(DBoard *board, XWindowAttributes *xwa)
{
  Display *d;
  Window w, child;
  int x, y;

  d = board->display;
  w = board->window;
  x = 0;
  y = 0;
  XMoveWindow(d, w, x, y);
}
#if 0
static void dBoard_set_window_pos(DBoard *board, XWindowAttributes *xwa)
{
  Display *d;
  Window w;

  /* sets window x,y,w,h values */
  XWindowChanges xwc;

  /* Adjust from window origin, to border origin */
  xwc.x = xwa->x - xwa->border_width - ch_offx;
  xwc.y = xwa->y - xwa->border_width - ch_offy;
  if (xwa->border_width == 0) {
    xwa->border_width = bwidth;
  }
  xwc.border_width = xwa->border_width;

  /* if we're less than max size in one axis, allow window manager doohickeys
     on the screen */
  if ((xwa->width < display_width) && (xwc.x < p_offx)) {
    xwc.x = p_offx;
  }
  if ((xwa->height < display_height) && (xwc.y < p_offy)) {
    xwc.y = p_offy;
  }
  /* Try to keep bottom right decorations inside */
  if ((xwc.x+width-debkludge_offx) > display_width) {
    xwc.x = display_width-width+debkludge_offx;
    if (xwc.x < 0) {
      xwc.x = 0;
    }
  }
  if ((xwc.y+height-debkludge_offy) > display_height) {
    xwc.y = display_height-height+debkludge_offy;
    if (xwc.y < 0) {
      xwc.y = 0;
    }
  }
  xwc.width = xwa->width;
  xwc.height = xwa->height;

  DEBUG_PRINT("SWP: xwa=%d,%d %dx%d xwc=%d,%d %dx%d off=%d,%d bw=%d klg=%d,%d\n",
              xwa->x, xwa->y, xwa->width, xwa->height,
              xwc.x, xwc.y, xwc.width, xwc.height, p_offx, p_offy,
              xwa->border_width, kludge_offx, kludge_offy);

  xwc.x += kludge_offx;
  xwc.y += kludge_offy;

  /* move/resize the window */
  XConfigureWindow(d, w,
                   CWX | CWY | CWWidth | CWHeight /*| CWBorderWidth*/, &xwc);
}
#endif /* 0 */

/*
 * set window and icon name.
 */
void dBoard_set_name(DBoard *board, const char *name)
{
  XrmValue value;
  char *winname, *iconname;
  char *title = TITLE_DEFAULT;
  Display *d;
  Window w;

  if (dBoard_resource(board, PROGNAME, ".title", NULL, &value)) {
    title = (char *)value.addr;
  }
  if (name[0] == '\0') {
    winname = XMALLOC(char, strlen(title)+1);
    iconname = XMALLOC(char, strlen(title)+1);
    sprintf(winname, "%s", title);
    sprintf(iconname, "%s", title);
  } else {
    winname = XMALLOC(char, strlen(title)+strlen(": ")+strlen(name)+1);
    iconname = XMALLOC(char, strlen(name)+1);
    sprintf(winname, "%s: %s", title, name);
    sprintf(iconname, "%s", name);
  }
  d = board->display;
  w = board->window;
  XStoreName(d, w, winname);
  XSetIconName(d, w, iconname);
  XFREE(winname);
  XFREE(iconname);
}

/*
 * Show window.
 */
void dBoard_show(DBoard *board)
{
  Display *d;
  Window w;
  Atom protocols;

  d = board->display;
  w = board->window;
  protocols = XInternAtom(d, "WM_DELETE_WINDOW", True);
  XSetWMProtocols(d, w, &protocols, 1);
  XMapWindow(d, w);
#if 0
  XWindowAttributes xwa;
  dBoard_get_window_pos(board, &xwa);
  dBoard_set_window_pos(board, &xwa);
  dBoard_get_window_pos(board, &xwa);
#endif /* 0 */
}

/*
 * close DBoard.
 */
int dBoard_close(DBoard *board)
{
  Display *d;
  Window w;
  int rc = TRUE;

  dBoard_free_cursor(board);
  d = board->display;
  w = board->window;
  XFreeGC(d, board->gc); /* free graphics contexts */
  XDestroyWindow(d, w);  /* delete window */
  XFlush(d);             /* flush */
  XrmDestroyDatabase(board->resource);
  XCloseDisplay(d);      /* close connection */
  XFREE(board);
  return rc;
}

/*
 * event loop.
 */
int dBoard_main(DBoard *board, void *data)
{
  int pending, done;
  XEvent event;
  Handler *handler;
  Processor *processor;

  handler = board->handler;
  processor = board->processor;
  pending = done = FALSE;
  while (!done) {
    pending = dBoard_pending(board);
    if (pending > 0) {
      dBoard_nextEvent(board, &event);
      (*handler)(board, &event, &done, data);
    } else {
      (*processor)(board, &done, data);
    }
  }
  return done;
}

/*
 * flushes the output buffer.
 */
int dBoard_flush(DBoard *board)
{
  return XFlush(board->display);
}

void dBoard_reset(DBoard *board, int width, int height)
{
  Display *display;
  Window window;
  int screen;
  int depth;
  Pixmap pixmap;

  display = board->display;
  window = board->window;
  screen = XDefaultScreen(display);
  depth  = XDefaultDepth(display, screen);
  pixmap = board->pixmap;
  if (pixmap != None) {
    XFreePixmap(display, pixmap);
  }
  pixmap = XCreatePixmap(display, window, width, height, depth);
  board->pixmap = pixmap;
  board->width  = width;
  board->height = height;
  XResizeWindow(display, window, width, height);
}

void dBoard_copy(DBoard *board)
{
  XCopyArea(board->display, board->pixmap, board->window, board->gc,
            0, 0, board->width, board->height, 0, 0);
}

/*
 * select input event.
 */
int dBoard_setHandler(DBoard *board, Handler *handler, long event_mask)
{
  board->handler = handler;
  return XSelectInput(board->display, board->window, event_mask);
}

/*
 * non-event process.
 */
void dBoard_setProcessor(DBoard *board, Processor *processor)
{
  board->processor = processor;
}

/*
 * define cursor.
 */
void dBoard_define_cursor(DBoard *board, enum cursorShape shape)
{
  switch(shape) {
  case CURSOR_WAIT:
    XDefineCursor(board->display, board->window, board->cursor[shape]);
    break;
  default:
    XUndefineCursor(board->display, board->window);
    break;
  }
}

/*
 * undefine cursor.
 */
void dBoard_undefine_cursor(DBoard *board)
{
  XUndefineCursor(board->display, board->window);
}

/*
 * initialize cursors.
 */
static void dBoard_init_cursor(DBoard *board)
{
  for (int i = 0; i < CURSOR_MAX; i++) {
    board->cursor[i] = None;
  }
  board->cursor[CURSOR_WAIT] = XCreateFontCursor(board->display, XC_watch);
}

/*
 * delete cursors.
 */
static void dBoard_free_cursor(DBoard *board)
{
  XUndefineCursor(board->display, board->window);
  for (int i = 0; i < CURSOR_MAX; i++) {
    XFreeCursor(board->display, board->cursor[i]);
  }
}

/*
 * initialize RGB colors.
 */
static void dBoard_init_color(DBoard *board)
{
  Display *display;
  Window window;
  int screen;
  int depth;
  unsigned long screen_set[SCREEN_RGB_CIMAX][SCREEN_RGB_IMAX];

  display = board->display;
  window = board->window;
  screen = XDefaultScreen(display);
  depth  = XDefaultDepth(display, screen);
  Pixmap pixmap = XCreatePixmap(display, window, 1, 1, depth);
  GC check_gc = XCreateGC(display, pixmap, 0, NULL);
  Colormap cmap = XDefaultColormap(display, screen);
  for (int ci = 0; ci < SCREEN_RGB_CIMAX; ci++) {
    for (int i = 0; i < SCREEN_RGB_IMAX; i++) {
      XColor color;
      color.red = color.green = color.blue = 0;
      ((unsigned short *)&color.red)[ci] = (unsigned short)((i << 8) | i);
      XAllocColor(display, cmap, &color);
      screen_set[ci][i] = (((unsigned short *)&color.red)[ci] >> 8) & 0xff;
      XSetForeground(display, check_gc, color.pixel);
      XDrawPoint(display, pixmap, check_gc, 0, 0);
      XImage *image;
      image = XGetImage(display, pixmap, 0, 0, 1, 1, AllPlanes, ZPixmap);
      if (image != NULL) {
        board->bits_per_pixel = image->bits_per_pixel;
        switch (board->bits_per_pixel) {
        case 8:
          board->screen_rgb[ci][i] = *(CARD8 *)image->data;
          break;
        case 16:
          board->screen_rgb[ci][i] = *(CARD16 *)image->data;
          break;
        case 24:
          board->screen_rgb[ci][i] =
            ((unsigned long)*(CARD8 *)image->data << 16) |
            ((unsigned long)*(CARD8 *)(image->data + 1) << 8) |
            (unsigned long)*(CARD8 *)(image->data + 2);
          break;
        case 32:
          board->screen_rgb[ci][i] = *(CARD32 *)image->data;
          break;
        }
        XDestroyImage(image);
      }
    }
  }
  XFreeGC(display, check_gc);
  XFreePixmap(display, pixmap);
}

/*
 * returns number of events.
 */
static int dBoard_pending(DBoard *board)
{
  return XPending(board->display);
}

/*
 * copies the first event.
 */
static int dBoard_nextEvent(DBoard *board, XEvent *event)
{
  return XNextEvent(board->display, event);
}

static XrmDatabase dBoard_get_xrmdb(Display *d, Window windowid)
{
  XrmDatabase res;

  /* don't use XResourceManagerString, as it is a snapshot of the
     string when the display was opened, and doesn't change */
  Atom resAtom = XInternAtom(d, "RESOURCE_MANAGER", True);
  if (resAtom != None) {
    Atom actType;
    int  rc, actFormat;
    unsigned long nitems, nleft;
    unsigned char *data;
    Window w;

    w = (windowid != 0) ? windowid : RootWindow(d, 0);
    rc = XGetWindowProperty(d, w, resAtom, 0L, 1L, False,
                            XA_STRING, &actType, &actFormat, &nitems, &nleft,
                            (unsigned char **)&data);
    if ((rc == Success) && (actType == XA_STRING) && (actFormat == 8)) {
      if ((nitems > 0) && (data != NULL)) {
        XFree(data);
      }
      w = (windowid != 0) ? windowid : RootWindow(d, 0);
      rc = XGetWindowProperty(d, w, resAtom, 0L, (long)((nleft+4+3)/4), False,
                              XA_STRING, &actType, &actFormat, &nitems, &nleft,
                              (unsigned char **)&data);
      if ((rc == Success) && (actType == XA_STRING) && (actFormat == 8) &&
          (data != NULL)) {
        res = XrmGetStringDatabase((char *)data);
        XFree(data);
        return res;
      }
    }
  }
  char *dpy = XResourceManagerString(d);
  if (dpy != NULL) {
    res = XrmGetStringDatabase(dpy);
  } else {
    char *xdefaults = "/.Xdefaults";
    char *filename;
    char *homedir = getenv("HOME");
    if (homedir == NULL) {
      struct passwd *pw;
      char *user = getenv("USER");
      if (user == NULL) {
        pw = getpwuid(getuid());
      } else {
        pw = getpwnam(user);
      }
      homedir = (pw != NULL) ? pw->pw_dir : "";
    }
    filename = XMALLOC(char, strlen(homedir)+strlen(xdefaults)+1);
    strcpy(filename, homedir);
    strcat(filename, xdefaults);
    res = XrmGetFileDatabase(filename);
    XFREE(filename);
  }
  return res;
}

static int dBoard_preparse_options(char **name, char **display, int *argc, char **argv)
{
  XrmDatabase db = NULL;
  XrmName name_list[3];
  XrmName class_list[3];
  XrmRepresentation type;
  XrmValue value;
  int num_options = 0;
  int argc_tmp;
  char **argv_tmp;

  argc_tmp = *argc;
  argv_tmp = XMALLOC(char *, argc_tmp);
  memcpy(argv_tmp, argv, sizeof(char *)*argc_tmp);
  num_options = sizeof(options)/sizeof(options[0]);
  XrmParseCommand(&db, options, num_options, ".", &argc_tmp, argv_tmp);
  XFREE(argv_tmp);
  name_list[0] = class_list[0] = XrmPermStringToQuark(".");
  name_list[2] = class_list[2] = NULLQUARK;
  if (name != NULL) {
    name_list[1] = XrmPermStringToQuark("name");
    if (XrmQGetResource(db, name_list, name_list, &type, &value)) {
      *name = value.addr;
    }
  }
  if (display != NULL) {
    name_list[1] = XrmPermStringToQuark("display");
    if (XrmQGetResource(db, name_list, name_list, &type, &value)) {
      *display = value.addr;
    }
  }
  XrmDestroyDatabase(db);
  return TRUE;
}

static void dBoard_parse_options(DBoard *board, int *argc, char **argv)
{
  XrmDatabase db;
  char progname[] = PROGNAME;
  int number_of_options;

  db = board->resource;
  number_of_options = sizeof(options)/sizeof(options[0]);
  XrmParseCommand(&db, options, number_of_options, progname, argc, argv);
}

static int dBoard_resource(DBoard *board, char *progname, char *name, char *class, XrmValue *value)
{
  XrmDatabase db;
  char *type;
  char *name_str;
  Bool rc;

  db = board->resource;
  name_str = XMALLOC(char, strlen(progname)+strlen(name)+1);
  strcpy(name_str, progname);
  strcat(name_str, name);
  rc = XrmGetResource(db, name_str, class, &type, value);
  if (rc) {
    fprintf(stderr,
            " '%s' // value len %d, addr %p '%s'\n",
            name, value->size, value->addr, value->addr);
  }
  XFREE(name_str);
  return rc;
}
