/* $Id$ */
/*
 * PNG decoding.
 */
#include <stdlib.h>
#include <stdio.h>
#include <png.h>
#include <setjmp.h>
#include "module.h"

int mod_read_PNG(Entry *entry, ImageSet *images);
static int init_png_create(png_structp *png_ptr, png_infop *info_ptr);
static int get_png(png_structp png_ptr, png_infop info_ptr, ImageSet *images, Entry *entry);
static int set_background(png_structp png_ptr, png_infop info_ptr,
                          int bit_depth, int color_type, Entry *entry);
static Image *get_image(png_structp png_ptr, png_infop info_ptr,
                        png_uint_32 width, png_uint_32 height, int color_type,
                        int pass);
static int set_imagetype(png_structp png_ptr, png_infop info_ptr,
                         png_uint_32 width, png_uint_32 height, int color_type,
                         Image **image_ptr, int *linesize_ptr);
static const char *get_comment(png_structp png_ptr, png_infop info_ptr);
static void my_error(png_structp png_ptr, png_const_charp message);
static void my_warning(png_structp png_ptr, png_const_charp message);
#if defined PNG_APNG_SUPPORTED
static int get_apng(png_structp png_ptr, png_infop info_ptr, ImageSet *images, Entry *entry);
static int get_dispose_op(png_byte dispose, png_uint_32 index);
static int get_blend_op(png_byte blend);
static int get_delay(png_uint_16 next_frame_delay_num, png_uint_16 next_frame_delay_den);
#endif /* PNG_APNG_SUPPORTED */
#if defined DEBUG
static void info_print(png_infop info_ptr);
#endif /* DEBUG */


int mod_read_PNG(Entry *entry, ImageSet *images)
{
  png_structp png_ptr = NULL;
  png_infop info_ptr = NULL;
  FILE *fp;
  int rc = FALSE;

  if ((fp = fopen(entry->filename, "rb")) == NULL) {
    fprintf(stderr, "can't open %s\n", entry->filename);
    return FALSE;
  }
  if (!init_png_create(&png_ptr, &info_ptr)) {
    return FALSE;
  }
  /* setjmp() must be called in every function that calls a PNG-reading
   * libpng function */
  if (setjmp(png_jmpbuf(png_ptr))) {
    png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
    fclose(fp);
    return FALSE;
  }
  /* I/O initialization */
  png_init_io(png_ptr, fp);

  png_read_info(png_ptr, info_ptr);  /* read all PNG info up to image data */
#if defined PNG_APNG_SUPPORTED
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_acTL)) {
    rc = get_apng(png_ptr, info_ptr, images, entry);
  } else {
    rc = get_png(png_ptr, info_ptr, images, entry);
  }
#else
  rc = get_png(png_ptr, info_ptr, images, entry);
#endif /* PNG_APNG_SUPPORTED */
  png_read_end(png_ptr, info_ptr);

  /* finish */
  png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
  fclose(fp);
  return rc;
}

static int init_png_create(png_structp *png_ptr, png_infop *info_ptr)
{
  *png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, png_voidp_NULL, my_error, my_warning);
  if (*png_ptr == NULL) {
    return FALSE;
  }
  *info_ptr = png_create_info_struct(*png_ptr);
  if (*info_ptr == NULL) {
    png_destroy_read_struct(png_ptr, png_infopp_NULL, png_infopp_NULL);
    return FALSE;
  }
  return TRUE;
}

#if defined PNG_APNG_SUPPORTED
static int get_apng(png_structp png_ptr, png_infop info_ptr, ImageSet *images, Entry *entry)
{
  png_uint_32 num_frames;
  png_uint_32 num_plays;
  png_uint_32 width, height;
  int bit_depth, color_type;

  num_frames = png_get_num_frames(png_ptr, info_ptr);
  num_plays = png_get_num_plays(png_ptr, info_ptr);
  png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
               NULL, NULL, NULL);
#if defined DEBUG
  fprintf(stderr, "png: APNG, num_frames=%d, num_plays=%d.\n",
          num_frames, num_plays);
  info_print(info_ptr);
#endif /* DEBUG */
  int pass = set_background(png_ptr, info_ptr, bit_depth, color_type, entry);
#if defined DEBUG
  info_print(info_ptr);
#endif /* DEBUG */
  imageSet_set_size(images, width, height);
  imageSet_set_num_plays(images, num_plays);
  for (png_uint_32 i = 0; i < num_frames; i++) {
    png_uint_32 next_frame_width;
    png_uint_32 next_frame_height;
    png_uint_32 next_frame_x_offset = 0;
    png_uint_32 next_frame_y_offset = 0;
    int delay = 0;
    int dispose_op = DISPOSE_OP_NONE;
    int blend_op = BLEND_OP_SOURCE;
    Image *image;

    png_read_frame_head(png_ptr, info_ptr);
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_fcTL)) {
      png_uint_16 next_frame_delay_num;
      png_uint_16 next_frame_delay_den;
      png_byte next_frame_dispose_op;
      png_byte next_frame_blend_op;
      png_get_next_frame_fcTL(png_ptr, info_ptr,
                              &next_frame_width, &next_frame_height,
                              &next_frame_x_offset, &next_frame_y_offset,
                              &next_frame_delay_num, &next_frame_delay_den,
                              &next_frame_dispose_op, &next_frame_blend_op);
      delay = get_delay(next_frame_delay_num, next_frame_delay_den);
      dispose_op = get_dispose_op(next_frame_dispose_op, i);
      blend_op = get_blend_op(next_frame_blend_op);
    } else {
      /* the first frame doesn't have an fcTL so it's expected to be hidden, 
       * but we'll extract it anyway */
      next_frame_width = png_get_image_width(png_ptr, info_ptr);
      next_frame_height = png_get_image_height(png_ptr, info_ptr);
    }
    image = get_image(png_ptr, info_ptr, next_frame_width, next_frame_height, color_type, pass);
    if (image == NULL) {
      return FALSE;
    }
    image_set_delay(image, delay);
    image_set_dispose_op(image, dispose_op);
    image_set_blend_op(image, blend_op);
    image_set_offset(image, next_frame_x_offset, next_frame_y_offset);
    char *comment;
    comment = get_comment(png_ptr, info_ptr);
    image_set_comment(image, comment);

    imageSet_add_image(images, image);
  }
  return TRUE;
}
#endif /* PNG_APNG_SUPPORTED */

static int get_png(png_structp png_ptr, png_infop info_ptr, ImageSet *images, Entry *entry)
{
  png_uint_32 width, height;
  int bit_depth, color_type;
  Image *image;
  const char *comment;

  png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
               NULL, NULL, NULL);
#if defined DEBUG
  info_print(info_ptr);
#endif /* DEBUG */
  int pass = set_background(png_ptr, info_ptr, bit_depth, color_type, entry);
#if defined DEBUG
  info_print(info_ptr);
#endif /* DEBUG */
  imageSet_set_size(images, width, height);

  image = get_image(png_ptr, info_ptr, width, height, color_type, pass);
  if (image == NULL) {
    return FALSE;
  }
  comment = get_comment(png_ptr, info_ptr);
  image_set_comment(image, comment);
  imageSet_add_image(images, image);
  return TRUE;
}

static int set_background(png_structp png_ptr, png_infop info_ptr,
                          int bit_depth, int color_type, Entry *entry)
{
  double gamma;

  if (bit_depth < 8) {
    png_set_packing(png_ptr);
  }
  if (png_get_gAMA(png_ptr, info_ptr, &gamma)) {
    png_set_gamma(png_ptr, entry->display_gamma, gamma);
  } else {
    png_set_gamma(png_ptr, entry->display_gamma, 0.45);
  }
  int have_imagebg = FALSE; /* FIXME!! */
  if (have_imagebg) {
    png_color_16 my_background;
    unsigned short imagebgR, imagebgG, imagebgB;

    if (bit_depth == 16) {
      my_background.red   = entry->background_red;
      my_background.green = entry->background_green;
      my_background.blue  = entry->background_blue;
      my_background.gray = entry->background_gray;
    } else {
      my_background.red   = (entry->background_red >> 8);
      my_background.green = (entry->background_green >> 8);
      my_background.blue  = (entry->background_blue >> 8);
      my_background.gray = entry->background_green;
    }
    png_set_background(png_ptr, &my_background,
                       PNG_BACKGROUND_GAMMA_SCREEN, 0, entry->display_gamma);
    if ((color_type == PNG_COLOR_TYPE_GRAY_ALPHA ||
         (color_type == PNG_COLOR_TYPE_GRAY &&
          png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))) &&
        (imagebgR != imagebgG || imagebgR != imagebgB)) {
      png_set_gray_to_rgb(png_ptr);
      png_set_expand(png_ptr);
      int gray_to_rgb = TRUE; /* FIXME!! */
    }
  } else {
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_bKGD)) {
      png_color_16p background;
      png_get_bKGD(png_ptr, info_ptr, &background);
      png_set_background(png_ptr, background,
                         PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
    } else {
      png_color_16 my_background;
      my_background.red = my_background.green = my_background.blue =
        my_background.gray = 0;
      png_set_background(png_ptr, &my_background,
                         PNG_BACKGROUND_GAMMA_SCREEN, 0, entry->display_gamma);
    }
  }

  if (bit_depth == 16) {
    png_set_strip_16(png_ptr);
  }
  if (color_type == PNG_COLOR_TYPE_GRAY ||
      color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
    png_set_expand(png_ptr);
  }
  int pass = png_set_interlace_handling(png_ptr);
  png_read_update_info(png_ptr, info_ptr);
  return pass;
}

static Image *get_image(png_structp png_ptr, png_infop info_ptr,
                        png_uint_32 width, png_uint_32 height, int color_type,
                        int pass)
{
  Image *image;

  /* set image type and buffer */
  int linesize;
  if (!set_imagetype(png_ptr, info_ptr, width, height, color_type, &image, &linesize)) {
    return NULL;
  }

  /* read png image */
  png_start_read_image(png_ptr);
  for (int j = 0; j < pass; j++) {
    unsigned char *p = image_get_data(image);
    for (int i = 0; i < height; i++) {
      png_read_row(png_ptr, p, png_bytep_NULL);
      p += linesize;
    }
  }
  return image;
}

static int set_imagetype(png_structp png_ptr, png_infop info_ptr,
                         png_uint_32 width, png_uint_32 height, int color_type,
                         Image **image_ptr, int *linesize_ptr)
{
  Image *image;
  unsigned char *data;
  int bufsize, linesize;
  png_colorp palette;
  int num_palette;

  image = image_new();
  if (image == NULL) {
    return FALSE;
  }
  image_set_size(image, width, height);
  switch (color_type) {
  case PNG_COLOR_TYPE_RGB:
  case PNG_COLOR_TYPE_RGB_ALPHA:
    linesize = 3*width;
    if (linesize/3 < width) {
      /* we know width > 0 */
      return FALSE;
    }
    image_set_type(image, TYPE_RGB);
    break;
  case PNG_COLOR_TYPE_GRAY:
  case PNG_COLOR_TYPE_GRAY_ALPHA:
    linesize = width;
    image_set_type(image, TYPE_INDEXED);
    image_set_num_colors(image, 256);
    for (int i = 0; i < 256; i++) {
      image_set_color(image, i, i, i, i, 0x00);
    }
    break;
  case PNG_COLOR_TYPE_PALETTE:
    linesize = width;
    image_set_type(image, TYPE_INDEXED);
    if (png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette)) {
      image_set_num_colors(image, num_palette);
      for (int i = 0; i < num_palette; i++) {
        image_set_color(image, i, palette[i].red, palette[i].green, palette[i].blue, 0x00);
      }
    }
    break;
  }
  bufsize = linesize*height;
  if (bufsize/linesize < height) {
    return FALSE;
  }
  data = (unsigned char *)calloc(bufsize, sizeof(unsigned char));
  if (data == NULL) {
    free(image);
    return FALSE;
  }
  image_set_data(image, data);
  *linesize_ptr = linesize;
  *image_ptr = image;
  return TRUE;
}

static const char *get_comment(png_structp png_ptr, png_infop info_ptr)
{
  char *comment = NULL;
  png_textp text = NULL;
  int num_text = 0;

  png_get_text(png_ptr, info_ptr, &text, &num_text);
  if (num_text > 0) {
    const char *sep = "::";
    const char *eot = "\n";
    size_t size = 1;
    for (int i = 0; i < num_text; i++) {
      size += strlen(text[i].key)+strlen(sep)+strlen(text[i].text)+strlen(eot);
    }
    comment = (char *)malloc(size);
    if (comment == NULL) {
      png_warning(png_ptr, "can't allocate comment string");
    } else {
      comment[0] = '\0';
      for (int i = 0; i < num_text; i++) {
        strcat(comment, text[i].key);
        strcat(comment, sep);
        strcat(comment, text[i].text);
        strcat(comment, eot);
      }
    }
  }
  return comment;
}

static void my_error(png_structp png_ptr, png_const_charp message)
{
  fprintf(stderr, "%s\n", message);
  longjmp(png_ptr->jmpbuf, 1);
}

static void my_warning(png_structp png_ptr, png_const_charp message)
{
  fprintf(stderr, "%s\n", message);
}

#if defined PNG_APNG_SUPPORTED
static int get_dispose_op(png_byte dispose, png_uint_32 index)
{
  int dispose_op = DISPOSE_OP_NONE;
  switch (dispose) {
  case PNG_DISPOSE_OP_NONE:
    dispose_op = DISPOSE_OP_NONE;
    break;
  case PNG_DISPOSE_OP_BACKGROUND:
    dispose_op = DISPOSE_OP_BACKGROUND;
    break;
  case PNG_DISPOSE_OP_PREVIOUS:
    dispose_op = (index != 0) ? DISPOSE_OP_PREVIOUS : DISPOSE_OP_BACKGROUND;
    break;
  }
  return dispose_op;
}

static int get_blend_op(png_byte blend)
{
  int blend_op = BLEND_OP_SOURCE;
  switch (blend) {
  case PNG_BLEND_OP_SOURCE:
    blend_op = BLEND_OP_SOURCE;
    break;
  case PNG_BLEND_OP_OVER:
    blend_op = BLEND_OP_OVER;
    break;
  }
  return blend_op;
}

static int get_delay(png_uint_16 next_frame_delay_num, png_uint_16 next_frame_delay_den)
{
  int delay;
  if (next_frame_delay_den == 0) {
    next_frame_delay_den = 100;
  }
  delay = next_frame_delay_num*1000/next_frame_delay_den;
  return delay;
}
#endif /* PNG_APNG_SUPPORTED */

#if defined DEBUG
static void info_print(png_infop info_ptr)
{
  char *type;
  switch(info_ptr->color_type) {
  case PNG_COLOR_TYPE_PALETTE:
    type = "palette";
    break;
  case PNG_COLOR_TYPE_GRAY:
    type = "gray";
    break;
  case PNG_COLOR_TYPE_GRAY_ALPHA:
    type = "gray+alpha";
    break;
  case PNG_COLOR_TYPE_RGB:
    type = "rgb";
    break;
  case PNG_COLOR_TYPE_RGB_ALPHA:
    type = "rgb+alpha";
    break;
  }
  fprintf(stderr, "png: %d bit %s [%dx%d]\n",
          info_ptr->bit_depth*info_ptr->channels,
          type, (int)info_ptr->width, (int)info_ptr->height);
}
#endif /* DEBUG */
#if 0
#include <lcms.h>

cmsHPROFILE get_color_profile(png_structp png_ptr, png_infop info_ptr)
{
  /* First try to see if iCCP chunk is present */
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_iCCP)) {
    DWORD len;
    char *profile;
    int  compression;
    char *name;
    png_get_iCCP(png_ptr, info_ptr, &name, &compression, &profile, &len);
    return cmsOpenProfileFromMem(profile, len);
  }

  /* Try to see if sRGB is specified, then use it */
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_sRGB)) {
    char *sRGB_file = "/usr/share/color/icc/sRGB.icm";

    if (access(sRGB_file, 0) == 0) {
      return cmsOpenProfileFromFile(sRGB_file, "r");
    }
  }

  /* Then, try to see if chromacities are defined. */
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_cHRM) &&
      png_get_valid(png_ptr, info_ptr, PNG_INFO_gAMA)) {
    cmsCIExyY white;
    cmsCIExyYTRIPLE primaries;

    png_get_cHRM(png_ptr, info_ptr,
                 &white.x, &white.y,
                 &primaries.Red.x, &primaries.Red.y,
                 &primaries.Green.x, &primaries.Green.y,
                 &primaries.Blue.x, &primaries.Blue.y);
    white.Y = primaries.Red.Y = primaries.Green.Y = primaries.Blue.Y  = 1.0;

    cmsHPROFILE profile;
    double file_gamma;
    LPGAMMATABLE gamma_table[3];

    png_get_gAMA(png_ptr, info_ptr, &file_gamma);
    gamma_table[0] = gamma_table[1] = gamma_table[2] =
      cmsBuildGamma(256, 1/file_gamma);
    if (!gamma_table[0]) {
      SignalError(-1, "Create gammatable: cannot allocate");
    }
    profile = cmsCreateRGBProfile(&white, &primaries, gamma_table);
    cmsFreeGamma(gamma_table[0]);
    return profile;
 }

  /* Only gamma is given, assume sRGB primaries?? */
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_gAMA)) {
    cmsCIExyY white =  {0.3127, 0.3290, 1.0}; /* D65 */
    cmsCIExyYTRIPLE primaries = {
      {0.6400, 0.3300, 1.0},
      {0.3000, 0.6000, 1.0},
      {0.1500, 0.0600, 1.0}
    };
    cmsHPROFILE profile;
    double file_gamma;
    LPGAMMATABLE gamma_table[3];

    png_get_gAMA(png_ptr, info_ptr, &file_gamma);
    gamma_table[0] = gamma_table[1] = gamma_table[2] =
      cmsBuildGamma(256, 1/file_gamma);
    if (!gamma_table[0]) {
      SignalError(-1, "Create gammatable: cannot allocate");
    }
    profile = cmsCreateRGBProfile(&white, &primaries, gamma_table);
    cmsFreeGamma(gamma_table[0]);
    return profile;
  }

  /* The image cannot be accurately described */
  return NULL;
}
#endif /* 0 */
