/* $Id$ */
#include <math.h>
#include "common.h"
#include "module.h"

struct image {
  unsigned char *data; /* image data */
  int xoff, yoff;      /* offset for image */
  int width, height;   /* image size */
  ImageType type;      /* type of image */
  int num_colors;      /* number of indexed colors */
  unsigned char r[256], g[256], b[256]; /* color map for indexed palette */
  unsigned char a[256]; /* alpha channel for indexed palette */
  int delay;           /* interframe delay [milli second] */
  int dispose_op;      /* dispose operation */
  int blend_op;        /* blend operation */
  const char *comment; /* comment text */
  struct image *next;  /* pointer to next image */
};
struct imageSet {
  Image *frames;
  int width, height;   /* images dimensions */
  int num_images;      /* number of images */
  int num_loop;        /* number of loop */
  int num_colors;      /* number of indexed colors */
  unsigned char r[256], g[256], b[256]; /* color map for indexed palette */
  unsigned char a[256]; /* alpha channel for indexed palette */
  unsigned long background;
};
struct reader {
  const char *key;    /* magic data */
  size_t key_len;     /* magic data length */
  const char *format; /* image format */
  int (*mod_read)(Entry *entry, ImageSet *); /* read function */
};
typedef struct reader Reader;

#if defined ENABLE_GIF
extern int mod_read_GIF(Entry *entry, ImageSet *images);
#endif /* ENABLE_GIF */
#if defined ENABLE_JPEG
extern int mod_read_JPEG(Entry *entry, ImageSet *images);
#endif /* ENABLE_JPEG */
#if defined ENABLE_PNG
extern int mod_read_PNG(Entry *entry, ImageSet *images);
#endif /* ENABLE_PNG */
extern int mod_read_null(Entry *entry, ImageSet *images);

static void imageSet_init(ImageSet *images);

static void imageSet_transform(ImageSet *images0, double fx, double fy);
static Image *image_transform(Image *image0, double fx, double fy);
static void image_interpolate(Image *image0, Image *image, int x, int y, double fx, double fy);
static double length(double a1, double a2, double b1, double b2);

int module_read_images(Entry *entry, ImageSet *images, const char *format)
{
  const Reader readers[] = {
#if defined ENABLE_GIF
    { "GIF",      3, "gif", mod_read_GIF },
#endif /* ENABLE_GIF */
#if defined ENABLE_JPEG
    { "\xff\xd8", 2, "jpg", mod_read_JPEG },
#endif /* ENABLE_JPEG */
#if defined ENABLE_PNG
    { "\x89PNG",  4, "png", mod_read_PNG },
#endif /* ENABLE_PNG */
    { NULL, 0, NULL, mod_read_null }
  };
  const Reader *reader;

  if (format == NULL) {
    FILE *fp;
    char buf[16];
    fp = fopen(entry->filename, "rb");
    if (fp == NULL) {
      return FALSE;
    }
    if (fread(buf, sizeof(char), 16, fp) != sizeof(char)*16) {
      return FALSE;
    }
    fclose(fp);
    reader = readers;
    while (reader->key != NULL) {
      if (memcmp(buf, reader->key, reader->key_len) == 0) {
        break;
      }
      reader++;
    }
    if (reader->key == NULL) {
      fprintf(stderr, "Unknown file type '%s'\n", entry->filename);
    }
  } else {
    reader = readers;
    while (reader->key != NULL) {
      if (strcmp(format, reader->format) == 0) {
        break;
      }
      reader++;
    }
    if (reader->key == NULL) {
      fprintf(stderr, "Unknown format type '%s'\n", format);
    }
  }
  imageSet_init(images);
  if (!(*(reader->mod_read))(entry, images)) {
    fprintf(stderr, "Can not read file '%s'\n", entry->filename);
    return FALSE;
  }
#if 0
    double fx, fy;
    fx = fy = 0.5;
    imageSet_transform(images, fx, fy);
#endif /* 0 */
  return TRUE;
}

#if 0
foo()
{
  double fx, fy;

  /* -maxpect */
  fx = ((double)entry->display_width)/images0->width;
  fy = ((double)entry->display_height)/images0->height;
  if (fx > fy) {
    fx = fy;
  }
  imageSet_transform(images, fx, fy);
}
#endif /* 0 */

ImageSet *imageSet_new(void)
{
  ImageSet *images;

  images = XMALLOC(ImageSet, 1);
  if (images != NULL) {
    imageSet_init(images);
  }
  return images;
}

void imageSet_free(ImageSet *images)
{
  Image *frame, *next;
  unsigned char *data;
  char *comment;
  int n;

  n = images->num_images;
  frame = images->frames;
  for (int i = 0; i < n; i++) {
    data = frame->data;
    comment = (char *)frame->comment;
    XFREE(data);
    XFREE(comment);
    next = frame->next;
    XFREE(frame);
    frame = next;
    DEBUG_PRINT("imageSet_free: %d\n", i);
  }
  images->frames = NULL;
  images->num_images = 0;
  images->num_loop = 0;
  images->num_colors = 0;
}

/*
 * add an image frame.
 */
void imageSet_add_image(ImageSet *images, Image *image)
{
  Image *p;

  p = images->frames;
  if (p == NULL) {
    images->frames = image;
    images->num_images++;
    return;
  }
  while (p->next != NULL) {
    p = p->next;
  }
  p->next = image;
  images->num_images++;
}

/*
 * gets number of images.
 */
int imageSet_get_nimages(ImageSet *images)
{
  return images->num_images;
}

/*
 * gets first image.
 */
Image *imageSet_get_first(ImageSet *images)
{
  return images->frames;
}

/*
 * gets background color.
 */
unsigned long imageSet_get_background(ImageSet *images)
{
  return 0;
}

/*
 * gets screen size.
 */
void imageSet_get_size(ImageSet *images, int *width, int *height)
{
  *width = images->width;
  *height = images->height;
}

/*
 * sets screen size.
 */
void imageSet_set_size(ImageSet *images, int width, int height)
{
  images->width = width;
  images->height = height;
}

/*
 * gets number of loop.
 */
int imageSet_get_num_plays(ImageSet *images)
{
  return images->num_loop;
}

/*
 * sets number of loop.
 */
void imageSet_set_num_plays(ImageSet *images, int num)
{
  images->num_loop = num;
}

/*
 * sets number of index colors.
 */
void imageSet_set_num_colors(ImageSet *images, int num)
{
  if (num > 0) {
    images->num_colors = num;
  }
}

/*
 * sets index colors.
 */
void imageSet_set_color(ImageSet *images, int index, unsigned char r, unsigned char g, unsigned char b, unsigned char a)
{
  if (index < images->num_colors) {
    images->r[index] = r;
    images->g[index] = g;
    images->b[index] = b;
    images->a[index] = a;
  }
}

static void imageSet_init(ImageSet *images)
{
  images->frames = NULL;
  images->width = 320;
  images->height = 200;
  images->num_images = 0;
  images->num_loop = 0;
  images->num_colors = 0;
}

/*
 * transform images0 -> images
 */
static void imageSet_transform(ImageSet *images0, double fx, double fy)
{
  ImageSet *images;
  Image *image, *image0;
  int i, n;

  images = imageSet_new();
  if (images == NULL) {
    /* error */
    return;
  }
  n = images0->num_images;
  image0 = images0->frames;
  for (i = 0; i < n; i++) {
    DEBUG_PRINT("image_transform(): %d\n", i);
    image = image_transform(image0, fx, fy);
    if (image == NULL) {
      break;
    }
    imageSet_add_image(images, image);
    image0 = image0->next;
  }
  images->width = (int)floor(images0->width*fx);
  images->height = (int)floor(images0->height*fy);
  images->num_loop = images0->num_loop;
  images->num_colors = 0;

  imageSet_free(images0);
  *images0 = *images;
  free(images);
}

Image *image_new(void)
{
  Image *image;

  image = XMALLOC(Image, 1);
  if (image == NULL) {
    return NULL;
  }
  image->data = NULL;
  image->xoff = 0;
  image->yoff = 0;
  image->width = 0;
  image->height = 0;
  image->type = TYPE_RGB;
  image->num_colors = 0;
  image->delay = 0;
  image->dispose_op = DISPOSE_OP_NONE;
  image->blend_op = BLEND_OP_SOURCE;
  image->comment = NULL;
  image->next = NULL;
  return image;
}
#if 0
static void image_free(Image *image)
{
  unsigned char *data;
  char *comment;

  data = image->data;
  comment = (char *)image->comment;
  XFREE(data);
  XFREE(comment);
  XFREE(image);
}
#endif /* 0 */

/*
 * gets image data.
 */
unsigned char *image_get_data(Image *image)
{
  return image->data;
}

/*
 * sets image data.
 */
void image_set_data(Image *image, unsigned char *data)
{
  image->data = data;
}

/*
 * gets image size.
 */
void image_get_size(Image *image, int *width, int *height)
{
  *width = image->width;
  *height = image->height;
}

/*
 * sets image size.
 */
void image_set_size(Image *image, int width, int height)
{
  image->width = width;
  image->height = height;
}

/*
 * gets image offset.
 */
void image_get_offset(Image *image, int *xoff, int *yoff)
{
  *xoff = image->xoff;
  *yoff = image->yoff;
}

/*
 * sets image offset.
 */
void image_set_offset(Image *image, int xoff, int yoff)
{
  image->xoff = xoff;
  image->yoff = yoff;
}

/*
 * gets number of index colors.
 */
int image_get_num_colors(Image *image)
{
  return image->num_colors;
}

/*
 * sets number of index colors.
 */
void image_set_num_colors(Image *image, int num)
{
  if (num > 0) {
    image->num_colors = num;
  }
}

/*
 * gets index colors.
 */
void image_get_color(Image *image, int index, unsigned char *r, unsigned char *g, unsigned char *b, unsigned char *a)
{
  if (index < image->num_colors) {
    *r = image->r[index];
    *g = image->g[index];
    *b = image->b[index];
    *a = image->a[index];
  }
}

/*
 * sets index colors.
 */
void image_set_color(Image *image, int index, unsigned char r, unsigned char g, unsigned char b, unsigned char a)
{
  if (index < image->num_colors) {
    image->r[index] = r;
    image->g[index] = g;
    image->b[index] = b;
    image->a[index] = a;
  }
}

/*
 * sets comment.
 */
void image_set_comment(Image *image, const char *comment)
{
  image->comment = comment;
}

/*
 * gets delay [m sec].
 */
int image_get_delay(Image *image)
{
  return image->delay;
}

/*
 * sets delay [m sec].
 */
void image_set_delay(Image *image, int delay)
{
  image->delay = delay;
}

/*
 * gets dispose operation.
 */
int image_get_dispose_op(Image *image)
{
  return image->delay;
}

/*
 * sets dispose operation.
 */
void image_set_dispose_op(Image *image, int dispose_op)
{
  image->dispose_op = dispose_op;
}

/*
 * gets blend operation.
 */
int image_get_blend_op(Image *image)
{
  return image->blend_op;
}

/*
 * sets blend operation.
 */
void image_set_blend_op(Image *image, int blend_op)
{
  image->blend_op = blend_op;
}

/*
 * gets image type.
 */
ImageType image_get_type(Image *image)
{
  return image->type;
}

/*
 * sets image type.
 */
void image_set_type(Image *image, ImageType type)
{
  image->type = type;
}

/*
 * next image.
 */
Image *image_next(Image *image)
{
  return image->next;
}

/*
 * transform image0 -> image
 */
static Image *image_transform(Image *image0, double fx, double fy)
{
  Image *image;
  int x, y, xoff, yoff, w, h;
  unsigned char *data;

  image = (Image *)malloc(sizeof(Image));
  if (image == NULL) {
    return NULL;
  }
  w = (int)floor(image0->width*fx);
  h = (int)floor(image0->height*fy);
  xoff = (int)floor(image0->xoff*fx);
  yoff = (int)floor(image0->yoff*fy);
  data = (unsigned char *)malloc(w*h*3);
  if (data == NULL) {
    return NULL;
  }
  image->data = data;
  image->width = w;
  image->height = h;
  for (y = 0; y < h; y++) {
    for (x = 0; x < w; x++) {
      image_interpolate(image0, image, x, y, fx, fy);
    }
  }
  image->xoff = xoff;
  image->yoff = yoff;
  image->type = TYPE_RGB;
  image->num_colors = 0;
  image->delay = image0->delay;
  image->comment = (image0->comment == NULL) ? NULL : strdup(image0->comment);
  image->next = NULL;
  return image;
}

/*
 * liner interpolation
 */
static void image_interpolate(Image *image0, Image *image, int x, int y, double fx, double fy)
{
  double dx, dy, inv_x1, inv_y1, inv_x2, inv_y2;
  int x1, y1, x2, y2;
  int xa, ya;
  double r, g, b, weight, weight_sum, lx, ly;
  unsigned char *data, *data0;
  int width, width0;

  dx = (fx-1)/2;
  dy = (fy-1)/2;
  inv_x1 = (x-dx)/fx;
  inv_y1 = (y-dy)/fy;
  inv_x2 = (x+1+dx)/fx;
  inv_y2 = (y+1+dy)/fy;
  x1 = (int)floor(inv_x1);
  y1 = (int)floor(inv_y1);
  x2 = (int)floor(inv_x2);
  y2 = (int)floor(inv_y2);
  data0 = image0->data;
  width0 = image0->width;
  weight_sum = 0;
  r = g = b = 0;
  for (ya = y1; ya <= y2; ya++) {
    if ((ya < 0) || (ya >= image0->height)) {
      continue;
    }
    ly = length(ya, ya+1, inv_y1, inv_y2);
    for (xa = x1; xa <= x2; xa++) {
      unsigned char c;

      if ((xa < 0) || (xa >= image0->width)) {
        continue;
      }
      lx = length(xa, xa+1, inv_x1, inv_x2);
      weight = lx*ly;
      weight_sum += weight;
      switch (image0->type) {
      case TYPE_INDEXED:
        c = *(data0+(width0*ya+xa));
        r += weight*image0->r[c];
        g += weight*image0->g[c];
        b += weight*image0->b[c];
        break;
      case TYPE_RGB:
        r += weight*(*(data0+(width0*ya+xa)*3));
        g += weight*(*(data0+(width0*ya+xa)*3+1));
        b += weight*(*(data0+(width0*ya+xa)*3+2));
        break;
      }
    }
  }
  r /= weight_sum;
  g /= weight_sum;
  b /= weight_sum;
  data = image->data;
  width = image->width;
  *(data+(width*y+x)*3) = (int)floor(r+0.5);
  *(data+(width*y+x)*3+1) = (int)floor(g+0.5);
  *(data+(width*y+x)*3+2) = (int)floor(b+0.5);
}

/*
 * length of bounds A and B.
 *
 * a1 < b1 < a2 < b2 ; l = a2-b1
 * a1 < b1 < b2 < a2 ; l = b2-b1
 * b1 < a1 < a2 < b2 ; l = a2-a1
 * b1 < a1 < b2 < a2 ; l = b2-a1
 */
static double length(double a1, double a2, double b1, double b2)
{
  double l = 0;

  if ((a1 <= b1) && (b1 <= a2) && (a2 <= b2)) {
    l = a2-b1;
  }
  if ((a1 <= b1) && (b1 <= a2) && (b2 <= a2)) {
    l = b2-b1;
  }
  if ((b1 <= a1) && (a1 <= a2) && (a2 <= b2)) {
    l = a2-a1;
  }
  if ((b1 <= a1) && (a1 <= b2) && (b2 <= a2)) {
    l = b2-a1;
  }
  return l;
}
