/* $Id$ */
#include <X11/keysym.h> /* XK_q */
#include <X11/Xproto.h> /* CARD8, CARD32 */
#include <X11/Xutil.h> /* XDestroyImage() */
#include "common.h"
#include "cache.h"
#include "dboard.h"
#include "module.h"
#include "basetime.h"

#define DISPLAY_GAMMA 2.20 /* default display gamma */

struct wrapper {
  int index;
  DBoard *board;
  Image *image;
};

struct player {
  ImageSet *images;
  Image *current;
  Cache *cache;
  int count;
  int loop_count;
  int running, running0;
  BaseTime *t0, *t;
  const char *format;
  int argc;
  const char **argv;
  int arg_index;
};
typedef struct player Player;

static int x_main(int argc, char **argv);
static int read_options(const char **format, int *running, int *maxpect, int *argc, char **argv);

static void player_init(Player *player, const char *format, int running, int argc, const char **argv, int arg_index);
static void player_free(Player *player);
static void player_image_next(Player *player);
static void player_arg_next(Player *player);
static void player_arg_previous(Player *player);
static int player_read_file(Player *player);
static void player_reset(Player *player);
static int player_has_multiframe(Player *player);

static int moving_process(DBoard *board, int *pdone, void *data);
static int handle_event(DBoard *board, XEvent *event, int *pdone, void *data);
static void board_setup(DBoard *board, Player *player);
static void image_setup(DBoard *board, Player *player);
static void image_dispose(DBoard *board, ImageSet *images, XImage *prev_frame, int dispose_op);

static unsigned int ximage_hashcode(const void *arg);
static void *ximage_get(const void *arg);
static int ximage_release(void *data);

static XImage *ximage_setup(DBoard *board, Image *image);
static void ximage_indexed(XImage *ximage, Image *image, int x0, int y0, unsigned long screen_rgb[][SCREEN_RGB_IMAX]);
static void ximage_rgb(XImage *ximage, Image *image, int x0, int y0, unsigned long screen_rgb[][SCREEN_RGB_IMAX]);

int main(int argc, char **argv)
{
  x_main(argc, argv);
  return EXIT_SUCCESS;
}

static int x_main(int argc, char **argv)
{
  DBoard *board;
  Player player;

  board = dBoard_open(&argc, argv);
  if (board == NULL) {
    fprintf(stderr, "Cannot open toolkit.\n");
    exit(EXIT_FAILURE);
  }
  const char *format = NULL;
  int running = FALSE;
  int maxpect = FALSE;
  read_options(&format, &running, &maxpect, &argc, argv);
  if (argc < 2) {
    /* usage */
    exit(EXIT_FAILURE);
  }
  player_init(&player, format, running, argc, argv, 1);

  if (!player_read_file(&player)) {
    fprintf(stderr, "Cannot read file.\n");
    player_free(&player);
    exit(EXIT_FAILURE);
  }
  board_setup(board, &player);
  image_setup(board, &player);
  dBoard_setHandler(board, handle_event,
                    ExposureMask |
                    KeyPressMask | KeyReleaseMask | ColormapChangeMask);
  dBoard_setProcessor(board, moving_process);
  dBoard_show(board);
  player_reset(&player);
  while (!dBoard_main(board, &player)) {
    ;
  }
  dBoard_close(board);
  player_free(&player);
  return TRUE;
}

static int read_options(const char **format, int *running, int *maxpect, int *argc, char **argv)
{
  char **argsave;
  int n;

  n = *argc;
  argsave = ++argv;
  n--;
  *format = NULL;
  while (n > 0) {
    if (strcmp(*argv, "-format") == 0) {
      argv++;
      (*argc)--;
      n--;
      if (n > 0) {
        *format = *argv;
        argv++;
        (*argc)--;
        n--;
      }
    }
    if (strcmp(*argv, "-a") == 0) {
     argv++;
      (*argc)--;
      n--;
      *running = TRUE;
    }
    *argsave = *argv;
    argsave++;
    argv++;
    n--;
  }
  return TRUE;
}

static void entry_init(Entry *entry, DBoard *board)
{
  unsigned long bg;
  int depth, w, h;

  dBoard_get_attributes(board, &depth, &w, &h, &bg);
  entry->filename = NULL;
  entry->display_gamma = DISPLAY_GAMMA;
  entry->background_red = (bg >> 16) & 0xff;
  entry->background_green = (bg >> 8) & 0xff;
  entry->background_blue = bg & 0xff;
  entry->background_gray = (entry->background_red+entry->background_green+entry->background_blue)/3;
  entry->bit_depth = depth;
  entry->display_width = w;
  entry->display_height = h;
}

static void player_init(Player *player, const char *format, int running, int argc, const char **argv, int arg_index)
{
  player->t0 = baseTime_new();
  player->t = baseTime_new();
  player->argc = argc;
  player->argv = argv;
  player->arg_index = 1;
  player->format = format;
  player->running0 = running;
  player->images = imageSet_new();
  player->cache = NULL;
}

static void player_free(Player *player)
{
  baseTime_delete(player->t0);
  baseTime_delete(player->t);
  imageSet_free(player->images);
  XFREE(player->images);
  cache_close(player->cache);
}

static void player_image_next(Player *player)
{
  Image *image;
  int num_loop;

  image = player->current;
  if (image_next(image) == NULL) {
    player->current = imageSet_get_first(player->images);
    player->count = 0;
    player->loop_count++;
  } else {
    player->current = image_next(image);
    player->count++;
  }
  num_loop = imageSet_get_num_plays(player->images);
  if (num_loop > 0) {
    if (player->loop_count >= num_loop) {
      player->running = FALSE;
    }
  }
}

static void player_arg_next(Player *player)
{
  if (++(player->arg_index) >= player->argc) {
    player->arg_index = player->argc-1;
  }
}

static void player_arg_previous(Player *player)
{
  if(--(player->arg_index) < 1) {
    player->arg_index = 1;
  }
}

static int player_read_file(Player *player)
{
  Entry entry;
  const char *filename;
  const char *format;
  int index, rc;

  index = player->arg_index;
  filename = (player->argv)[index];
  format = player->format;
  DEBUG_PRINT("filename = '%s'\n", filename);
  entry.display_gamma = DISPLAY_GAMMA;
  entry.background_red = 0;
  entry.background_green = 0;
  entry.background_blue = 0;
  entry.background_gray = 0;
  entry.bit_depth = 0;
  entry.filename = filename;
  cache_close(player->cache);
  imageSet_free(player->images);
  rc = module_read_images(&entry, player->images, format);
  player_reset(player);
  if (!rc) {
    return FALSE;
  }
  return TRUE;
}

static void player_reset(Player *player)
{
  player->current = imageSet_get_first(player->images);
  player->count = 0;
  player->loop_count = 0;
  if (player_has_multiframe(player)) {
    player->running = player->running0;
    player->cache = cache_open(imageSet_get_nimages(player->images), ximage_hashcode, ximage_get, ximage_release);
  } else {
    player->running = FALSE;
    player->cache = NULL;
  }
  baseTime_get(player->t0);
}

static int player_has_multiframe(Player *player)
{
  return (imageSet_get_nimages(player->images) > 1);
}

static int moving_process(DBoard *board, int *pdone, void *data)
{
  Player *player;
  BaseTime *t;
  int dt, delay;

  player = (Player *)data;
  if (!player->running) {
    baseTime_sleep_msec(20);
    return TRUE;
  }
  delay = image_get_delay(player->current);
  t = player->t;
  baseTime_get(t);
  dt = baseTime_subtract_msec(t, player->t0);
  if (dt >= delay) {
    player_image_next(player);
    image_setup(board, player);
    dBoard_copy(board);
    baseTime_get(t);
    dt = baseTime_subtract_msec(t, player->t0);
    baseTime_set(player->t0, t);
    DEBUG_PRINT("dt = %d, delay = %d\n", dt, delay);
  } else {
    int wait = (delay-dt)/10;
    if (wait < 20) {
      baseTime_sleep_msec(wait);
    } else {
      baseTime_sleep_msec(20);
    }
  }
  return TRUE;
}

static int handle_event(DBoard *board, XEvent *event, int *pdone, void *data)
{
  Player *player;
  int rc, done;
  Atom delwin;
  KeySym keysym;

  player = (Player *)data;
  rc = done = FALSE;
  switch (event->type) {
  case Expose:
    DEBUG_PRINT("Expose\n");
    if ((event->xexpose).count == 0 ) {
      dBoard_copy(board);
      dBoard_flush(board);
    }
    break;
  case NoExpose:
    DEBUG_PRINT("NoExpose\n");
    break;
  case ClientMessage:
    DEBUG_PRINT("ClientMessage\n");
    delwin = XInternAtom(board->display, "WM_DELETE_WINDOW", True);
    if ((event->xclient).data.l[0] == delwin) {
      done = TRUE;
    }
    break;
  case KeyRelease:
    keysym = XLookupKeysym(&(event->xkey), 0);
    if (keysym == XK_n) {
      if (player_has_multiframe(player)) {
        player_image_next(player);
        image_setup(board, player);
        dBoard_copy(board);
        dBoard_flush(board);
      }
    } else if (keysym == XK_space) {
      dBoard_define_cursor(board, CURSOR_WAIT);
      dBoard_flush(board);
      player_arg_next(player);
      player_read_file(player);
      board_setup(board, player);
      image_setup(board, player);
      dBoard_copy(board);
      dBoard_flush(board);
      dBoard_undefine_cursor(board);
    } else if (keysym == XK_BackSpace) {
      dBoard_define_cursor(board, CURSOR_WAIT);
      dBoard_flush(board);
      player_arg_previous(player);
      player_read_file(player);
      board_setup(board, player);
      image_setup(board, player);
      dBoard_copy(board);
      dBoard_flush(board);
      dBoard_undefine_cursor(board);
    } else if (keysym == XK_s) {
      if (player_has_multiframe(player)) {
        player->running = !(player->running);
      }
    } else if (keysym == XK_q) {
      done = TRUE;
    } else {
      ;
    }
    break;
  case DestroyNotify:
    DEBUG_PRINT("DestroyNotify\n");
    done = TRUE;
    break;
  default:
    DEBUG_PRINT("event: %d\n", event->type);
    break;
  }
  *pdone = done;
  return rc;
}

static void board_setup(DBoard *board, Player *player)
{
  int width, height;
  int index;
  const char *filename;

  imageSet_get_size(player->images, &width, &height);
  dBoard_reset(board, width, height);
  index = player->arg_index;
  filename = (player->argv)[index];
  dBoard_set_name(board, filename);
}

struct previousFrame {
  XImage *frame;
  int dispose_op;
};

static void image_setup(DBoard *board, Player *player)
{
  Display *display;
  GC gc;
  Pixmap pixmap;
  XImage *ximage;
  XImage *last_frame;
  Image *image;
  int width, height;
  int xoff, yoff;
  int dispose_op;
  unsigned long xcolor;
  XImage *prev_frame;
  int prev_dispose_op;

  DEBUG_PRINT(" current count = %d\n", player->count);
  image = player->current;
  if (image == NULL) {
    return;
  }
  display = board->display;
  pixmap = board->pixmap;
  last_frame = NULL;
  dispose_op = image_get_dispose_op(image);
  if (dispose_op == DISPOSE_OP_PREVIOUS) {
    int w, h;
    imageSet_get_size(player->images, &w, &h);
    last_frame = XGetImage(display, pixmap, 0, 0, w, h, AllPlanes, ZPixmap);
  }
  image_dispose(board, player->images, prev_frame, prev_dispose_op);
  prev_frame = last_frame;
  prev_dispose_op = dispose_op;
  if (player->cache != NULL) {
    struct wrapper t;

    t.index = player->count;
    t.board = board;
    t.image = image;
    ximage = (XImage *)cache_get(player->cache, &t);
  } else {
    ximage = ximage_setup(board, image);
  }
  image_get_size(image, &width, &height);
  image_get_offset(image, &xoff, &yoff);
  gc = board->gc;
  XPutImage(display, pixmap, gc, ximage, 0, 0, xoff, yoff, width, height);
  if (player->cache == NULL) {
    XDestroyImage(ximage);
  }
}

static void image_dispose(DBoard *board, ImageSet *images, XImage *prev_frame, int dispose_op)
{
  Display *display;
  Pixmap pixmap;
  GC gc;
  int width, height;
  unsigned long xcolor;

  display = board->display;
  pixmap = board->pixmap;
  gc = board->gc;
  switch (dispose_op) {
  case DISPOSE_OP_NONE:
    break;
  case DISPOSE_OP_BACKGROUND:
    imageSet_get_size(images, &width, &height);
    xcolor = imageSet_get_background(images);
    XSetForeground(display, gc, xcolor);
    XFillRectangle(display, pixmap, gc, 0, 0, width, height);
    break;
  case DISPOSE_OP_PREVIOUS:
    imageSet_get_size(images, &width, &height);
    XPutImage(display, pixmap, gc, prev_frame, 0, 0, 0, 0, width, height);
    break;
  default:
    DEBUG_PRINT("unknown dipose_op type %d. why?\n", dispose_op);
    break;
  }
}

static unsigned int ximage_hashcode(const void *arg)
{
  struct wrapper *p;

  p = (struct wrapper *)arg;
  return p->index;
}

static void *ximage_get(const void *arg)
{
  struct wrapper *p;

  p = (struct wrapper *)arg;
  return ximage_setup(p->board, p->image);
}

static int ximage_release(void *p)
{
  return XDestroyImage((XImage *)p);
}

static XImage *ximage_setup(DBoard *board, Image *image)
{
  Display *display;
  Pixmap pixmap;
  XImage *ximage;
  ImageType type;
  int width, height;
  int xoff, yoff, x0, y0;
  int full_image = FALSE;

  if (full_image) {
    /*imageSet_get_size(images, &width, &height);*/
    image_get_offset(image, &x0, &y0);
    xoff = 0;
    yoff = 0;
  } else {
    image_get_size(image, &width, &height);
    image_get_offset(image, &xoff, &yoff);
    x0 = 0;
    y0 = 0;
  }
  type = image_get_type(image);
  display = board->display;
  pixmap = board->pixmap;
  ximage = XGetImage(display, pixmap, xoff, yoff, width, height, AllPlanes, ZPixmap);
  DEBUG_PRINT("image type: %d %d (%d, %d) [%dx%d]\n",
              type, image_get_num_colors(image),
              xoff, yoff, width, height);
  switch (type) {
  case TYPE_RGB:
    ximage_rgb(ximage, image, x0, y0, board->screen_rgb);
    break;
  case TYPE_INDEXED:
    ximage_indexed(ximage, image, x0, y0, board->screen_rgb);
    break;
  default:
    fprintf(stderr, " unknown image type %d. why?\n", type);
    break;
  }
  return ximage;
}

static void ximage_indexed(XImage *ximage, Image *image, int x0, int y0, unsigned long screen_rgb[][SCREEN_RGB_IMAX])
{
  unsigned long xcolors[SCREEN_RGB_IMAX];
  unsigned char alpha[SCREEN_RGB_IMAX];
  unsigned char *pp;
  int w, h, x, y, n;

  n = image_get_num_colors(image);
  for (int i = 0; i < n; i++) {
    unsigned char r, g, b, a;
    image_get_color(image, i, &r, &g, &b, &a);
    xcolors[i] = screen_rgb[0][r] | screen_rgb[1][g] | screen_rgb[2][b];
    alpha[i] = a;
  }
  pp = image_get_data(image);
  image_get_size(image, &w, &h);
  y = y0;
  for (int j = 0; j < h; j++) {
    x = x0;
    for (int i = 0; i < w; i++) {
      if (alpha[*pp] < 0xff) {
        XPutPixel(ximage, x, y, xcolors[*pp]);
      }
      pp++;
      x++;
    }
    y++;
  }
}

static void ximage_rgb(XImage *ximage, Image *image, int x0, int y0, unsigned long screen_rgb[][SCREEN_RGB_IMAX])
{
  unsigned long xcol;
  unsigned char *pp;
  int w, h, x, y;

  pp = image_get_data(image);
  image_get_size(image, &w, &h);
  y = y0;
  for (int j = 0; j < h; j++) {
    x = x0;
    for (int i = 0; i < w; i++) {
      xcol  = screen_rgb[0][*pp++];
      xcol |= screen_rgb[1][*pp++];
      xcol |= screen_rgb[2][*pp++];
      XPutPixel(ximage, i, j, xcol);
      x++;
    }
    y++;
  }
}
